#!/usr/bin/perl

require '6502cmd.pl';

$FILE = $ARGV[0];
$saddr = 49152;

for($pass=0;$pass<2;$pass++) {
	$addr=$saddr;
	$line=0;
	open(FILE,$FILE);
	while(<FILE>) {
		$line++;
		if ( /^[ \t]*ORG[ \t](.+)/ ) {
			$addr = $1;
		} elsif ( /^[ \t]*SET[ \t]+(.+)[ \t]+(.+)/ ) {
			$label{$1}=$2 if (!$pass);
		} elsif ( /^;/ ) {
			next;
		} elsif ( /^[ \t]*MEM[ \t]+(.+)/ ) {
			if ($pass) {
				for($i=0;$i<$1;$i++) {
					push @mem,0;
				}
			}
			$addr+=$1;
		} elsif ( /^[ \t]*STR "(.+)"/ ) {
			if ($pass) {
				push @mem,unpack('C*',$1);
				push @mem,0;
			}
			$addr+=length($1)+1;
		} elsif ( /^(\w+):$/ ) {
			$label{$1}=$addr if (!$pass);
		} elsif ( /^[ \t]*(\w+)[ \t]*$/ ) {
			implied($1) if $pass;
			$addr++;
		} elsif ( /^[ \t]*(\w+)[ \t]+\[(.+)\]/ ) {
			relative($1,$2) if $pass;
			$addr+=2;
		} elsif ( /^[ \t]*(\w+)[ \t]+(.+),X/ ) {
			absolute_x($1,$2) if $pass;
			$addr+=3;
		} elsif ( /^[ \t]*(\w+)[ \t]+(.+),Y/ ) {
			absolute_y($1,$2) if $pass;
			$addr+=3;
		} elsif ( /^[ \t]*(\w+)[ \t]+#(.+)/ ) {
			immediate($1,$2) if $pass;
			$addr+=2;
		} elsif ( /^[ \t]*(\w+)[ \t]+(.+)/ ) {
			absolute($1,$2) if $pass;
			$addr+=3;
		}
	
	}
	close(FILE);
}

($lo,$hi) = decomp($saddr);

# This line puts the destination address at the start of the code,
# so that a Commodore 64 knows where to put it when loading it in.

printf("%c%c",$lo,$hi);

foreach $i (@mem) {
	printf("%c",$i);
}

exit(0);

sub implied {
	my($cmd)=@_;
	$code = $cmd{"$cmd:IMPLIED"};
	push @mem,$code;
}

sub immediate {
	my($cmd,$arg)=@_;
	if (defined($cmd{"$cmd:IMMEDIATE"})) {
		$code = $cmd{"$cmd:IMMEDIATE"};
		push @mem,$code,parsearg($arg);
	} elsif (defined($cmd{"$cmd:RELATIVE"})) {
		$code = $cmd{"$cmd:RELATIVE"};
		push @mem,$code,parsearg($arg);
	}
}

sub relative {
	my($cmd,$arg)=@_;
		$code = $cmd{"$cmd:RELATIVE"};
		$arg=parsearg($arg);
		if ($arg > $addr) {
			$arg = $arg - $addr - 2;
			error("Branch out of range") if ($arg > 127);
		} else {
			$arg = 255 - (1 + $addr - $arg);
			error("Branch out of range") if ($arg < 128);
			
		}
		push @mem,$code,$arg;
}

sub error {
	my($s)=@_;
	printf STDERR "$s at line $line\n";
	exit(1);
}

sub absolute {
	my($cmd,$arg)=@_;
	if (defined($cmd{"$cmd:ABSOLUTE"})) {
		$code = $cmd{"$cmd:ABSOLUTE"};
		if (defined($label{$arg})) {
			$arg = $label{$arg};
		}
		my($lo,$hi)=decomp($arg);
		push @mem,$code,$lo,$hi;
	} elsif (defined($cmd{"$cmd:RELATIVE"})) {
		$code = $cmd{"$cmd:RELATIVE"};
		if (defined($label{$arg})) {
			$arg = $label{$arg};
		}
		if ($arg > $addr) {
			$arg = $arg - $addr - 2;
		}
		push @mem,$code,$arg;
	}
}

sub absolute_x {
	my($cmd,$arg)=@_;
	$code = $cmd{"$cmd:ABSOLUTE_X"};
	if (defined($label{$arg})) {
		$arg = $label{$arg};
	}
	my($lo,$hi)=decomp($arg);
	push @mem,$code,$lo,$hi;
}

sub absolute_y {
	my($cmd,$arg)=@_;
	$code = $cmd{"$cmd:ABSOLUTE_Y"};
	if (defined($label{$arg})) {
		$arg = $label{$arg};
	}
	my($lo,$hi)=decomp($arg);
	push @mem,$code,$lo,$hi;
}

sub decomp {
	my($arg)=@_;
	my($lo,$hi);

	$hi = int ($arg / 256) ;
	$lo = $arg - 256 * $hi;
	return ($lo,$hi);
}

sub parsearg {
	my($arg)=@_;
	my($getlo,$gethi);
	my($lo,$hi);

	# Traditionally, C64 assembler used '$' to represent hex.
	# Converts '$' to'0x'. The assembler will also allow users to
	# use '0x'

	$arg =~ s/\$/0x/;

	if (substr($arg,0,1) eq '<') {
		$getlo=1;
		$arg=substr($arg,1);
	}
	if (substr($arg,0,1) eq '>') {
		$gethi=1;
		$arg=substr($arg,1);
	}
	if (defined($label{$arg})) {
		$arg = $label{$arg};
	}
	if ($getlo || $gethi) {
		my($lo,$hi)=decomp($arg);
		return $lo if ($getlo);
		return $hi if ($gethi);
	}

	# Perlism. Add zero to make sure hex numbers become integers.
	return ($arg+0);
}
